/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/app.cpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Application class implementation.
**/

#include "app.hpp"

#include <argp.h>

#if ENABLE_LIBINPUT
    #include "interceptor-li.hpp"
#endif // ENABLE_LIBINPUT
#if ENABLE_GNOME
    #include "layouter-gnome.hpp"
#endif // ENABLE_GNOME
#if ENABLE_X
    #include "interceptor-x.hpp"
    #include "layouter-x.hpp"
#endif // ENABLE_X
#include "posix.hpp"

// Definition of argp variables must be in global namespace.

/**
    Program version string. It is printed in respond to `--version` option. The string may include
    copyright.
**/
char const * argp_program_version =
    PACKAGE_STRING ".\n"
    "Copyright (C) " PACKAGE_COPYRIGHT_YEAR " " PACKAGE_COPYRIGHT_HOLDER ".\n"
    "License " PACKAGE_LICENSE " <" PACKAGE_LICENSE_URL ">.\n"
    "This is free software: you are free to change and redistribute it." "\n"
    "There is NO WARRANTY, to the extent permitted by law."
    ;

/** Email address to send bug reports to. May be `nullptr`. **/
char const * argp_program_bug_address = nullptr; /*
    argp separates "Report bugs to" from preceeding text with empty line. I do not like it.
    Let's reset bug report address, I'll include it into help text manually.
*/

/** Status to return in case of command line error. **/
error_t argp_err_exit_status = tapper::app_t::cmdline_error;

namespace tapper {

// -------------------------------------------------------------------------------------------------
// app_t
// -------------------------------------------------------------------------------------------------

static void handler( int signal ) {
    DBG( "Signal " << signal );
};

#if ! ENABLE_GNOME
    namespace Glib {
        /**
            Fake Glib::Exception class to let code be compiled successfully if Gnome support is
            disabled.
        **/
        class Exception : public std::exception {
        };
    };
#endif // ENABLE_GNOME


static strings_t const default_assignments {
    "29=1",     // Left Ctrl
    "97=2",     // Right Ctrl
};

app_t::app_t(
    int     argc,
    char *  argv[]
) :
    OBJECT_T(),
    prog_name( program_invocation_short_name )
{
    /*
        Glibmm may throw an exception. Unfortunately, `Glib::Exception` does not inherit
        `std::exception`, so `main` will complain on unknown exception. Let's convert
        `Glib::Exception` to `app_t::error_t` to avoid leaking internal details.
    */
    try {

        posix::signal::set_t    set( { SIGINT, SIGTERM } );
        posix::signal::action_t action( handler, set );
        posix::signal::action( SIGINT,  action );
        posix::signal::action( SIGTERM, action );

        parse_argv( argc, argv );

        switch ( _action ) {
            case action_t::run : {
                run();
            } break;
            case action_t::save_settings : {
                save_settings();
            } break;
            case action_t::list_keys : {
                list_keys();
            } break;
            case action_t::list_layouts : {
                list_layouts();
            } break;
            default : {
                assert( 0 );
            };
        }; // switch

    } catch ( Glib::Exception const & ex ) {

        ERR( ex.what() );

    };
}; // ctor

app_t::~app_t(
) {
    /*
        `_interceptor` and `_layouter` will be reset automatically, but I want to print "Done"
        after destructing them, not before.
    */
    _interceptor.reset();
    _layouter.reset();
    DBG( "Done." );
};

enum opt_t {
    // Short options:
    opt_x = 'x',
    // Long options:
    opt_bell = 1000,
    opt_gnome_shell,
    opt_interceptor,
    opt_layouter,
    opt_libinput,
    opt_list_keys,
    opt_list_layouts,
    opt_load_settings,
    opt_no_bell,
    opt_no_load_settings,
    opt_save_settings,
    opt_show_taps,
    opt_quiet,
    opt_xkb,
    opt_xrecord,
}; // enum opt_t

/**
    Parses command line option or argument. The method is static because it is called from within
    `argp_parse`.
**/
::error_t
app_t::parse_opt_or_arg(
    int             key,        ///< Option key (a short name or integer id).
    char *          _arg,       ///< String to parse.
    argp_state *    state
) {

    char const * arg = _arg; /*
        It is a little trick. I am not going to change argument, but parameter should be of `char
        *` type.
    */

    auto app = static_cast< app_t * >( state->input );

    // cppcheck-suppress shadowFunction
    THIS( app );

    try {       // No exceptions should leak from this method.

        switch ( key ) {

            case ARGP_KEY_INIT : {
            } break;

            case opt_bell : {
                app->set_bell( settings_t::bell_t::enabled );
            } break;

            case opt_gnome_shell : {
                if ( not ENABLE_GNOME ) {
                    ERR( "Program is built without Gnome support." );
                }; // if
                app->set_layouter( settings_t::layouter_t::gnome_shell );
            } break;

            case opt_interceptor : {
                app->set_interceptor( settings_t::to_interceptor( arg ) );
            } break;

            case opt_layouter : {
                app->set_layouter( settings_t::to_layouter( arg ) );
            } break;

            case opt_libinput : {
                if ( not ENABLE_LIBINPUT ) {
                    ERR( "Program is built without libinput support." );
                };
                app->set_interceptor( settings_t::interceptor_t::libinput );
            } break;

            case opt_list_keys : {
                app->set_action( action_t::list_keys );
            } break;

            case opt_list_layouts : {
                app->set_action( action_t::list_layouts );
            } break;

            case opt_load_settings : {
                if ( not ENABLE_GNOME ) {
                    ERR( "Program is built without Gnome support." );
                };
                app->_load_settings = true;
            } break;

            case opt_no_bell : {
                app->set_bell( settings_t::bell_t::disabled );
            } break;

            case opt_no_load_settings : {
                app->_load_settings = false;
            } break;

            case opt_save_settings : {
                if ( not ENABLE_GNOME ) {
                    ERR( "Program is built without Gnome support." );
                };
                app->set_action( action_t::save_settings );
            } break;

            case opt_show_taps : {
                app->_show_taps = true;
            } break;

            case opt_quiet : {
                app->_quiet = true;
            } break;

            case opt_x : {
                if ( not ENABLE_X ) {
                    ERR( "Program is built without X Window System support." );
                };
                app->set_interceptor( settings_t::interceptor_t::xrecord );
                app->set_layouter( settings_t::layouter_t::xkb );
            } break;

            case opt_xkb : {
                if ( not ENABLE_X ) {
                    ERR( "Program is built without X Window System support." );
                };
                app->set_layouter( settings_t::layouter_t::xkb );
            } break;

            case opt_xrecord : {
                if ( not ENABLE_X ) {
                    ERR( "Program is built without X Window System support." );
                };
                app->set_interceptor( settings_t::interceptor_t::xrecord );
            } break;

            case ARGP_KEY_NO_ARGS : {
                // There are no arguments, so settings were not loaded yet. Load them now:
                if ( app->_load_settings ) {
                    app->load_settings();
                };
                if ( app->_settings.assignments.empty() ) {
                    for ( auto & str : default_assignments ) {
                        arg = str.c_str();  // In case of problem code below will use `arg`.
                        app->parse_arg( arg );
                    };
                };
            } break;

            case ARGP_KEY_ARG : {
                /*
                    Load settings just before processing the first argument, because settings (e.
                    g. interceptor) may affect argument parsing.
                */
                if ( app->_load_settings and state->arg_num == 0 ) {
                    app->load_settings();
                };
                app->parse_arg( arg );
            } break;

            case ARGP_KEY_SUCCESS : {
            } break;

            case ARGP_KEY_ERROR : {
            } break;

            case ARGP_KEY_END : {
                app->_used_keys.clear();
            } break;

            case ARGP_KEY_FINI : {
            } break;

            default : {
                return ARGP_ERR_UNKNOWN;
            } break;

        }; // switch

    } catch ( std::exception const & ex ) {

        char const * format = nullptr;
        switch ( key ) {
            case ARGP_KEY_ARG : {
                format = "Bad argument %s: %s";
            } break;
            case ARGP_KEY_NO_ARGS : {
                format =
                    "Oops, bad default argument %s: %s\n"
                    "Try to specify explicit argument(s) to avoid this problem.";
            } break;
            default : {
                format = "Bad option %s: %s";
                arg = state->argv[ state->next - 1 ];
            } break;
        }; // switch
        argp_error( state, format, q( arg ).c_str(), ex.what() );

    }; // try

    return 0;

}; // parse_opt_or_arg

/** Parses entire command line. **/
void
app_t::parse_argv(
    int     argc,
    char *  argv[]
) {

    int const doc_option   = OPTION_DOC | OPTION_NO_USAGE;
    int const gnome_option = ENABLE_GNOME    ? 0 : OPTION_HIDDEN;
    int const x_option     = ENABLE_X        ? 0 : OPTION_HIDDEN;
    int const li_option    = ENABLE_LIBINPUT ? 0 : OPTION_HIDDEN;

    string_t const interceptor_help =
        "Interceptor to use, one of: " + join( ", ", settings_t::interceptors() );
    string_t const layouter_help =
        "Layouter to use, one of: " + join( ", ", settings_t::layouters() );

    argp_option opts[] = {
        { "Interceptor selection:", 0,                  nullptr, doc_option,
            "",
            100 },
        { "interceptor",            opt_interceptor,    "ARG",   0,
            interceptor_help.c_str(),
            101 },
        { "libinput",               opt_libinput,       nullptr, li_option,
            "Same as --interceptor=libinput",
            102 },
        { "xrecord",                opt_xrecord,        nullptr, x_option,
            "Same as --interceptor=xrecord",
            103 },

        { "Layouter selection:",    0,                  nullptr, doc_option,
            "",
            200 },
        { "layouter",               opt_layouter,       "ARG",   gnome_option,
            layouter_help.c_str(),
            201 },
        { "gnome-shell",            opt_gnome_shell,    nullptr, gnome_option,
            "Same as --layouter=gnome-shell",
            202 },
        { "xkb",                    opt_xkb,            nullptr, x_option,
            "Same as --layouter=xkb",
            203 },

        { "Bell mode:",             0,                  nullptr, doc_option,
            "",
            300 },
        { "bell",                   opt_bell,           nullptr, 0,
            "Ring a bell when a layout is activated",
            301 },
        { "no-bell",                opt_no_bell,        nullptr, 0,
            "Do not ring a bell when a layout is activated",
            302 },

        { "Settings management:",   0,                  nullptr, doc_option | gnome_option,
            "",
            400 },
        { "load-settings",      opt_load_settings,      nullptr, gnome_option,
            "Load settings (interceptor, layouter, bell mode, "
                "key/layout assignments)",
            401 },
        { "no-load-settings",   opt_no_load_settings,   nullptr, gnome_option,
            "Do not load settings",
            402 },
        { "save-settings",      opt_save_settings,      nullptr, gnome_option,
            "Save settings and exit",
            403 },

        { "Other options:",     0,                      nullptr, doc_option,
            "",
            500 },
        { "quiet",              opt_quiet,              nullptr, 0,
            "Do not print introductory messages",
            501 },
        { "x",                  opt_x,                  nullptr, x_option,
            "Same as --xrecord --xkb",
            502 },

        { "Help options:",      0,                      nullptr, doc_option,
            "",
            -2  },
        { "list-keys",          opt_list_keys,          nullptr, 0,
            "List available keys and exit",
            -1  },
        { "list-layouts",       opt_list_layouts,       nullptr, 0,
            "List available layouts and exit",
            -1  },
        { "show-taps",          opt_show_taps,          nullptr, 0,
            "Print key code of every tapped key",
            -1  },

        { 0 }

    }; // opts

    strings_t const disabled = {
        #if ! ENABLE_GNOME
            "Gnome",
        #endif
        #if ! ENABLE_LIBINPUT
            "libinput",
        #endif
        #if ! ENABLE_X
            "X Window System",
        #endif
    };

    string_t const doc = STR(
        "\n"
        PACKAGE_NAME " — " PACKAGE_SUMMARY ". "
        PACKAGE_NAME " works in background, "
        "when KEY is tapped, " PACKAGE_NAME " activates the corresponding keyboard LAYOUT. "
        "The assigned key always activates the same layout: "
        PACKAGE_NAME " never toggles layouts. "
        "\n\n"
        "Options:"
        "\v"    // Options will be listed here.
        "KEY is either a key code (an unsigned integer) or a symbolic key name. "
        "Key codes are always libinput key codes regardless of the interceptor. "
        "Key names depend on interceptor.\n"
        "\n"
        "\"Tap\" means a single key is pressed and quickly released. "
        "If a key is released slowly, "
        "or pressed/released in a combination with other key(s), it is not a tap.\n"
        "\n"
        "LAYOUT is either layout number (an unsigned integer) or layout name. "
        "Layouts are layouter-dependent.\n"
        "\n"
        << (
            ENABLE_GNOME ?
                "By default " PACKAGE_NAME " load its configuration "
                "(interceptor and layouter selections, bell mode, key/layout assignments) "
                "from GSettings database. "
                "Command line options and assignments override settings.\n"
                "\n"
            :
                ""
        ) <<
        "Exit status:\n"
        "  " << ok << " if everything is ok,\n"
        "  " << cmdline_error << " if usage (command-line) error,\n"
        "  " << runtime_error << " if runtime error.\n"
        "\n"
        "See " << prog_name << "(1) man page for more details.\n"
        << (
            disabled.empty() ?
            "" :
            PACKAGE_NAME " is built without support for " + join( ", ", disabled ) + ".\n"
        ) <<
        PACKAGE_NAME " home page: <" PACKAGE_URL ">.\n"
        "Report bugs to: <" PACKAGE_BUGREPORT ">."
    );

    argp parser = { opts, parse_opt_or_arg, "[KEY=[LAYOUT]...]", doc.c_str() };
    ::error_t error = argp_parse( & parser, argc, argv, 0, 0, this );
    if ( error ) {
        exit( argp_err_exit_status );
    }; // if

}; // parse_argv

/**
    Parses one argument. All arguments should follow "KEY=LAYOUT" format, where both KEY and LAYOUT
    can be either name or number of key and layout respectively.
**/
void
app_t::parse_arg(
    string_t const & arg
) {
    auto parts  = split( '=', arg, 2 );
    auto key    = parse_key( parts[ 0 ] );
    auto layout = parse_layout( parts.size() > 1 ? parts[ 1 ] : "" );
    DBG( "key " << interceptor().key_name( key ) << " => layout " << layout_name( layout ) );
    // Check if the key is already assigned:
    auto inserted = _used_keys.insert( key );
    if ( not inserted.second ) {    // key was not inserted because it already is in the set.
        ERR( "Key " << interceptor().key_name( key ) << " has been already assigned." );
    }; // if
    // Save the assignment in the settings:
    if ( layout ) {
        _settings.assignments.insert( { key, layout } );
    } else {
        _settings.assignments.erase( key );     // Cancel key assignment, if any.
    };
}; // parse_arg

key_t
app_t::parse_key(
    string_t const & str
) {
    // TODO: Allow key in form "50:LFSH"?
    key_t key = 0;
    if ( str.empty() ) {
        ERR( "No key specified." );
    }; // if
    if ( like_integer( str ) ) {    // Looks like a key code.
        try {
            key = parse_int( str );
            auto & range = interceptor().key_range();
            if ( not range.includes( key ) ) {
                ERR( "Out of range " << range << "." );
            }; // if
        } catch ( error_t const & ex ) {
            ERR( "Bad key code: " << ex.what() );
        };
    } else {                        // Looks like a key name.
        key = interceptor().key( str );
        if ( not key ) {
            ERR( "Bad key name: Unknown name." );
        };
    }; // if
    return key;
}; // parse_key

uint_t
app_t::parse_layout(
    string_t const & layout
) {
    uint_t code = 0;
    if ( like_integer( layout ) ) {
        try {
            code = parse_int( layout );
            uint_t min = 1;
            uint_t max = layouter().layout_max();
            if ( code and not ( min <= code and code <= max ) ) {
                ERR( "Out of range [" << min << ".." << max << "]." );
            }; // if
        } catch ( error_t const & ex ) {
            ERR( "Bad layout number: " << ex.what() );
        };
    } else {
        auto name = layout;
        if ( not name.empty() ) {
            auto names = layouter().layouts();
            auto it    = find( name, names );
            // TODO: What if layout names are not unique?
            if ( it == npos ) {
                ERR( "Bad layout name: Unknown name." );
            }; // if
            code = it + 1;
        };
    }; // if
    return code;
}; // parse_layout

uint_t
app_t::parse_int(
    string_t const & str
) {
    int num = 0;
    try {
        num = val< decltype( num ) >( str );
    } catch ( val_error_t const & ex ) {
        ERR( ex.what() );
    };
    if ( num < 0 ) {
        ERR( "Negative number is not allowed." );
    };
    return num;
};

void
app_t::set_interceptor(
    settings_t::interceptor_t interceptor
) {
    if ( _settings.interceptor != settings_t::interceptor_t::unset ) {
        ERR( "Interceptor has already been set." );
    };
    _settings.interceptor = interceptor;
};

void
app_t::set_layouter(
    settings_t::layouter_t layouter
) {
    if ( _settings.layouter != settings_t::layouter_t::unset ) {
        ERR( "Layouter has already been set." );
    };
    _settings.layouter = layouter;
};

void
app_t::set_bell(
    settings_t::bell_t bell
) {
    if ( _settings.bell != settings_t::bell_t::unset ) {
        ERR( "Bell has already been set." );
    };
    _settings.bell = bell;
};

void
app_t::set_action(
    action_t action
) {
    if ( _action != action_t::dflt ) {
        ERR( "Action has already been set." );
    };
    _action = action;
};

void
app_t::load_settings(
) {
    /*
        Settings extracted from the command line have priority over settings loaded from GSettings.
    */
    _settings = settings_t::load() + _settings;
};

void
app_t::save_settings(
) {
    _settings.save();
};

/** Prints list of known keys. **/
void
app_t::list_keys(
) {
    auto & key_names = interceptor().key_names();
    if ( key_names.empty() ) {
        WRN(
            interceptor().type() << " interceptor does not provide any key names; "
                << "use key codes in range " << interceptor().key_range() << "."
        );
    } else {
        typedef std::map< key_t, strings_t > key_codes_t;
        key_codes_t codes;
        for ( auto it : key_names ) {
            auto inserted = codes.insert( { it.second, { it.first } } );
            if ( not inserted.second ) {
                inserted.first->second.push_back( it.first );
            };
        }; // for
        for ( auto it : codes ) {
            for ( auto jt : it.second ) {
                OUT( it.first << ":" << jt );
            };
        };
    };
}; // list_keys

/** Prints list of known layouts. **/
void
app_t::list_layouts(
) {
    size_t n = 0;
    for ( auto & name : layouter().layouts() ) {
        ++ n;
        if ( not name.empty() ) {
            OUT( n << ":" << name );
        }; // if
    }; // for
}; // list_layouts

/** Runs Tapper in normal mode. **/
void
app_t::run(
) {
    /*
        Settings' assignments use 1-based layout numbers (layout #0 is a special case to "unassign"
        a previously assigned key). But tapper's assignments should be 0-based.
    */
    settings_t::assignments_t assignments;
    for ( auto const & assignment : _settings.assignments ) {
        if ( assignment.second ) {
            assignments[ assignment.first ] = assignment.second - 1;
        };
    };
    if ( ! _quiet ) {
        print_intro();
    };
    tapper_t tapper( interceptor(), layouter(), assignments, _show_taps );
    tapper.start();
    posix::sleep();
    tapper.stop();
};

/**
    Print introductionary information: version, interceptor and layouter selection, bell mode,
    and key/layout assignments.
**/
void
app_t::print_intro(
) {
    OUT( PACKAGE_STRING "." );
    OUT( "Starting in " << interceptor().type() << "/" << layouter().type() << " mode." );
    OUT( "Bell is " << settings_t::to_string( bell() ) << "." );
    // I want the assignments are sorted by layouts, not keys.
    typedef settings_t::assignments_t assignments_t;
    std::map< assignments_t::mapped_type, std::vector< assignments_t::key_type > > map;
    for ( auto const & it : _settings.assignments ) {
        map[ it.second ].push_back( it.first );
    };
    // And now print the map:
    for ( auto const & it : map ) {
        string_t keys;
        for ( size_t i = 0, end = it.second.size(); i < end; ++ i ) {
            keys +=
                ( i == 0 ? "" : ( i + 1 < end ? ", " : " or " ) ) +
                interceptor().key_name( it.second[ i ] );
        };
        OUT( "Tap key " << keys << " " << "to select layout " << layout_name( it.first ) << "." );
    }; // for
}; // print_intro

string_t
app_t::layout_name(
    uint_t code
) {
    auto names = layouter().layouts();
    auto name  = code < names.size() ? names[ code - 1 ] : "";
    return STR( code << ( name.empty() ? "" : ":" + name ) );
}; // layout_name

interceptor_t &
app_t::interceptor(
) {
    if ( not _interceptor ) {
        interceptor_t * interceptor { nullptr };
        if ( _settings.interceptor <= settings_t::interceptor_t::Auto ) {
            auto const type = posix::get_env( "XDG_SESSION_TYPE" );
            if ( ENABLE_X and type == "x11" ) {
                _settings.interceptor = settings_t::interceptor_t::xrecord;
            } else if ( ENABLE_LIBINPUT ) {
                _settings.interceptor = settings_t::interceptor_t::libinput;
            };
        };
        switch ( _settings.interceptor ) {
            #if ENABLE_X
                case settings_t::interceptor_t::xrecord : {
                    interceptor = new interceptor::x_t();
                } break;
            #endif // ENABLE_X
            #if ENABLE_LIBINPUT
                case settings_t::interceptor_t::libinput : {
                    interceptor = new interceptor::li_t();
                } break;
            #endif // ENABLE_LIBINPUT
            default : {
                ERR( "No interceptors." );
            };
        };
        _interceptor.reset( interceptor );
    };
    return * _interceptor.get();
};

layouter_t &
app_t::layouter(
) {
    if ( not _layouter ) {
        layouter_t * layouter { nullptr };
        if ( _settings.layouter <= settings_t::layouter_t::Auto ) {
            auto const desktop = posix::get_env( "XDG_CURRENT_DESKTOP" );
            if ( ENABLE_GNOME and desktop == "GNOME" ) {
                _settings.layouter = settings_t::layouter_t::gnome_shell;
            } else if ( ENABLE_X ) {
                _settings.layouter = settings_t::layouter_t::xkb;
            };
        };
        switch ( _settings.layouter ) {
            #if ENABLE_X
                case settings_t::layouter_t::xkb : {
                    layouter = new layouter::x_t();
                } break;
            #endif // ENABLE_X
            #if ENABLE_GNOME
                case settings_t::layouter_t::gnome_shell : {
                    layouter = new layouter::gnome_t();
                } break;
            #endif // ENABLE_GNOME
            default : {
                ERR( "No layouters." );
            } break;
        };
        _layouter.reset( layouter );
        _layouter->set_bell( _settings.bell == settings_t::bell_t::enabled );
    };
    return * _layouter.get();
};

settings_t::bell_t
app_t::bell(
) {
    if ( _settings.bell == settings_t::bell_t::unset ) {
        _settings.bell = settings_t::bell_t::disabled;
    };
    return _settings.bell;
};

}; // namespace tapper

// end of file //
