/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/app.hpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Application class interface.
**/

#ifndef _APP_HPP_
#define _APP_HPP_

#include <set>

#include "base.hpp"
#include "settings.hpp"
#include "interceptor.hpp"
#include "layouter.hpp"
#include "tapper.hpp"

struct argp_state;  // instead of #include <argp.h>.

namespace tapper {

class app_t : public object_t {

    public :

        /** Application exit status. **/
        enum exit_status_t {
            ok,                 ///< Everything is ok.
            not_ok,             ///< False, but not an error. Not actually used.
            cmdline_error,      ///< Usage error.
            runtime_error,      ///< Runtime error.
        };

        enum class action_t {
            dflt,
            run = dflt,
            save_settings,
            list_keys,
            list_layouts,
        };

        class error_t : public std::runtime_error {
            public :
                using std::runtime_error::runtime_error;
        }; // class error_t

    public :

        app_t( int argc, char * argv[] );
        ~app_t();

    private :

        typedef ptr_t< strings_t >  strings_p;

        // Command line parsing routines:
        static ::error_t parse_opt_or_arg( int key, char * _arg, argp_state * state );
        void   parse_argv( int argc, char * argv[] );
        void   parse_arg( string_t const & arg );
        key_t  parse_key( string_t const & str );
        uint_t parse_layout( string_t const & layout );
        uint_t parse_int( string_t const & str );

        // Low-level actions, called from parsing routines:
        void set_interceptor( settings_t::interceptor_t interceptor );
        void set_layouter( settings_t::layouter_t layouter );
        void set_bell( settings_t::bell_t bell );
        void set_action( action_t action );
        void load_settings();
        void save_settings();

        // Top-level actions:
        void list_keys();
        void list_layouts();
        void run();

        void print_intro();

        string_t            layout_name( uint_t code );

        interceptor_t &     interceptor();
        layouter_t &        layouter();
        settings_t::bell_t  bell();

    private :

        char const * const  prog_name;

        settings_t          _settings;
        bool                _load_settings { ENABLE_GNOME };
        bool                _show_taps { false };
        bool                _quiet { false };
        action_t            _action { action_t::dflt };
        std::set< key_t >   _used_keys;

        interceptor_p       _interceptor;
        layouter_p          _layouter;

}; // class app_t

}; // namespace tapper

#endif // _APP_HPP_

// end of file //
