/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/base.hpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

#ifndef _BASE_HPP_
#define _BASE_HPP_

#include "config.h"

#include <exception>
#include <map>              // std::map
#include <memory>           // std::unique_ptr

#include <cassert>
#include <cstddef>          // size_t
#include <cstdint>          // uint32_t

#include "reverse.hpp"
#include "string.hpp"
#include "ui.hpp"

namespace tapper {

using std::size_t;

typedef std::uint32_t   uint_t;
template< typename obj_t > using ptr_t = std::unique_ptr< obj_t >;

/**
    Integer type to represent event's time. Could be either 32- or 64-bit integer.
**/
typedef uint_t  time_t;     // TODO: uint64_t?

/**
    Key code. Represented by `unt_t`, but actual range of key codes is expected to be much
    narrower, something like [1..255]. See `key_max` in `interceptor_t`.
**/
typedef uint_t key_t;

struct key_range_t {
    key_range_t( uint_t _min = 1, uint_t _max = 255 );

    bool includes( uint_t code ) const;

    key_range_t & operator +=( key_t key );
    key_range_t & operator +=( key_range_t const & that );

    key_t  min;
    key_t  max;
}; // struct key_range_t

typedef ptr_t< key_range_t > key_range_p;

typedef std::map< string_t, key_t >  key_names_t;
typedef ptr_t< key_names_t >         key_names_p;
typedef reverse_map_t< key_names_t > key_codes_t;
typedef ptr_t< key_codes_t >         key_codes_p;

std::ostream &
operator <<(
    std::ostream & stream,
    key_range_t    range
);

/**
    Base class for all other classes.

    Objects are not copyable.

    It provides debugging facilities: each object has type (object type is used for constructing
    object id but is not saved as data member), number, and id.

    Object type is a string, it should be specified in the constructor. It is expected `OBJECT_T()`
    is used for it:

    @code {.cpp}
        class class_t : public object_t {
            public :
                class_t() : OBJECT_T() {
                    ...
                };
                ...
        };
    @endcode

    Object number is assigned automatically, sequentially within the type. The first created object
    of the type receives no 1, the next created object of the same type receives number 2, and so
    on. Numbers of deleted objects are not reused.

    Object id is constructed from object type and object number, and looks like "object#1".

    All these stuff is used for debugging purposes solely.

**/
class object_t {

    // Objects are non-copyable.
    object_t( object_t const & ) = delete;
    object_t & operator =( object_t const & ) = delete;

    #if defined( DEBUG )

        public :

            explicit object_t( char const * type );
            virtual ~object_t();

            object_t const * _this_() const;
            uint_t           _no_()   const;
            string_t         _id_()   const;

        private :

            uint_t const    _no;    ///< Object number, unique within the type.
            string_t const  _id;    ///< Object id, unique within the program.
    #else

        public :

            explicit object_t( char const * type ) {};
            virtual ~object_t() = default;

    #endif

}; // object_t

/**
    `ERR`, `WRN`, and `DBG` macro rely on `_this_()` method or global function. However, it does not
    work in static methods since `_this_` method is not accessible. `THIS` macro provides `_this_`
    definition for static methods:

    @code
    void a_static_method() {
        THIS( pointer-to-this-object );
        ...
        ERR( "message" );
    };
    @endcode
**/
#define THIS( OBJ ) auto _this_ = [ OBJ ] () { return OBJ; }; static_cast< void >( _this_ );

#if defined( DEBUG )
    #define OBJECT_T()  object_t( __FUNCTION__ )
    object_t const * _this_();
    void _dbg( object_t const * obj, char const * file, int line, string_t const & mesg );
    #define _DBG( _OBJ_, _ARGS_ ) _dbg( _OBJ_, __FILE__, __LINE__, STR( _ARGS_ ) )
#else
    #define OBJECT_T()  object_t( nullptr )
    #define _DBG( _OBJ_, _ARGS_ ) void( 0 )
#endif

#define _WRN( _OBJ_, _ARGS_ ) {                                                                    \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    eprint( severity_t::warning, _MSG_ );                                                          \
}
#define _ERR( _OBJ_, _ARGS_, ... ) {                                                               \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    _DBG( _OBJ_, "Throwing: " << _MSG_ );                                                          \
    throw error_t( _MSG_, ## __VA_ARGS__ );                                                        \
}
#define DBG( _ARGS_ )       _DBG( _this_(), _ARGS_ )
#define INF( _ARGS_ )       eprint( severity_t::info, STR( _ARGS_ ) )
#define WRN( _ARGS_ )       _WRN( _this_(), _ARGS_ )
#define ERR( _ARGS_, ... )  _ERR( _this_(), _ARGS_, ## __VA_ARGS__ )
#define OUT( _ARGS_ )       oprint( STR( _ARGS_ ) )

#define _CATCH_ALL( _OBJ_, _STATEMENT_ ) {                                                         \
    try {                                                                                          \
        _STATEMENT_;                                                                               \
    } catch ( std::exception const & _ex_ ) {                                                      \
        _WRN( _OBJ_, _ex_.what() );                                                                \
    } catch ( ... ) {                                                                              \
        _WRN( _OBJ_, "Unknown exception occurred" );                                               \
    };                                                                                             \
}

#define CATCH_ALL( _STATEMENT_ ) \
    _CATCH_ALL( _this_(), _STATEMENT_ )

#if defined( DEBUG )

    class _func_tracer_t {
        public :
            _func_tracer_t( object_t const * obj, char const * file, int line, char const * func );
            ~_func_tracer_t();
        private :
            object_t const * const  _obj;
            char const * const      _file;
            int const               _line;
            char const * const      _func;
    }; // class _func_tracer_t

    #define _TRACE( _OBJ_ ) \
        _func_tracer_t _tracer_ ## __COUNTER__( _OBJ_, __FILE__, __LINE__, __FUNCTION__ )
    #define TRACE() \
        _TRACE( _this_() )
#else

    #define _TRACE( _OBJ_ )
    #define TRACE()

#endif

}; // namespace tapper

#endif // _BASE_HPP_

// end of file //
