/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/interceptor-li.cpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

#include "interceptor-li.hpp"

#include <limits>

#if HAVE_LINUX_INPUT_EVENT_CODES_H
    #include "linux/input-event-codes.h"
#else
    #define KEY_RESERVED 0
    #define KEY_MAX      255
#endif // HAVE_LINUX_INPUT_EVENT_CODES_H

namespace tapper {
namespace interceptor {

// -------------------------------------------------------------------------------------------------
// li_t
// -------------------------------------------------------------------------------------------------

li_t::li_t(
) :
    OBJECT_T(),
    _context( std::bind( & li_t::_on_intercept, this, std::placeholders::_1 ), "seat0" )
{
}; // ctor

li_t::~li_t(
) {
}; // dtor

string_t
li_t::type(
) {
    return "libinput";
}; // type

key_range_t const &
li_t::key_range(
) {
    return _range;
}; // key_range

key_names_t const &
li_t::key_names(
) {
    static key_names_t const names {
        #if HAVE_LINUX_INPUT_EVENT_CODES_H
            #define X( NAME ) { #NAME, NAME },
            #include "input-event-names.def"
            #undef X
        #endif // HAVE_LINUX_INPUT_EVENT_CODES_H
    };
    return names;
}; // key_names

key_codes_t const &
li_t::key_codes(
) {
    static key_codes_t const codes {
        #if HAVE_LINUX_INPUT_EVENT_CODES_H
            #define X( NAME ) { NAME, #NAME },
            #include "input-event-names.def"
            #undef X
        #endif // HAVE_LINUX_INPUT_EVENT_CODES_H
    };
    return codes;
}; // key_codes

key_t
li_t::key(
    string_t const & name
) {
    auto & names = key_names();
    auto it = names.find( uc( name ) );
    if ( it == names.end() ) {
        it = names.find( "KEY_" + uc( name ) );
        if ( it == names.end() ) {
            return 0;
        };
    };
    return it->second;
}; // key

void
li_t::start(
) {
    _context.enable();
}; // start

void
li_t::stop(
) {
    _context.disable();
}; // stop

string_t
li_t::_key_name(
    key_t key
) {
    auto & codes = key_codes();
    auto it = codes.find( key );
    if ( it == codes.end() ) {
        return "";
    };
    return it->second;
}; // _key_name

void
li_t::_on_intercept(
    libinput_event * event
) {
    switch ( libinput_event_get_type( event ) ) {
        case LIBINPUT_EVENT_KEYBOARD_KEY : {
            auto kbev  = libinput_event_get_keyboard_event( event );
            auto time  = libinput_event_keyboard_get_time( kbev );
            auto key   = libinput_event_keyboard_get_key( kbev );
            auto state = libinput_event_keyboard_get_key_state( kbev );
            if ( _range.includes( key ) ) {
                _on_event( {
                    .time  = time,
                    .key   = key_t( key ),
                    .state = state == LIBINPUT_KEY_STATE_PRESSED ?
                        key_state_t::pressed : key_state_t::released,
                } );
            } else {
                WRN(
                    "libinput reported key " << key
                        << " out of expected range " << _range << "."
                );
            };
        } break;
        case LIBINPUT_EVENT_POINTER_BUTTON : {
            auto ptev   = libinput_event_get_pointer_event( event );
            auto time   = libinput_event_pointer_get_time( ptev );
            auto button = libinput_event_pointer_get_button( ptev );
            auto state  = libinput_event_pointer_get_button_state( ptev );
            if ( _range.includes( button ) ) {
                _on_event( {
                    .time  = time,
                    .key   = key_t( button ),
                    .state = state == LIBINPUT_BUTTON_STATE_PRESSED ?
                        key_state_t::pressed : key_state_t::released,
                } );
            } else {
                WRN(
                    "libinput reported button " << button
                        << " out of expected range " << _range << "."
                );
            };
        } break;
        default : {
        };
    };
}; // on_intercept

static_assert( KEY_RESERVED == 0 );
key_range_t const li_t::_range( 1, KEY_MAX );   // Do not include 0 to range.

}; // namespace interceptor
}; // namespace tapper

// end of file //
