/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/li.cpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    libinput wrappers implementation.
**/

#include "li.hpp"

#include <fcntl.h>
#include <unistd.h>
#include <libudev.h>
#include <libinput.h>
#include <poll.h>

#include <iostream>

#include "string.hpp"

namespace tapper {
namespace li {

// -------------------------------------------------------------------------------------------------
// class udev_t
// -------------------------------------------------------------------------------------------------

udev_t::udev_t(
) :
    rep( udev_new() )
{
    if ( not rep ) {
        ERR( "Failed to initialize udev" );
    };
};

udev_t::udev_t(
    myself_t const & that
) :
    rep( that.rep )
{
    udev_ref( rep );
};

udev_t::~udev_t(
) {
    udev_unref( rep );
    rep = nullptr;
};

udev_t::myself_t &
udev_t::operator =(
    myself_t const & that
) {
    udev_unref( rep );
    rep = that.rep;
    udev_ref( rep );
    return * this;
};

// -------------------------------------------------------------------------------------------------
// class context_t
// -------------------------------------------------------------------------------------------------

static
int
_open(
    char const *    path,
    int             flags,
    void *          data
) {
    return reinterpret_cast< context_t * >( data )->open( path, flags );
};

static
void
_close(
    int             fd,
    void *          data
) {
    return reinterpret_cast< context_t * >( data )->close( fd );
};

static
void _log_handler(
    struct libinput *           libinput,
    enum libinput_log_priority  priority,
    const char *                format,
    va_list                     args
) {
    char buffer[ 1024 ];
    // TODO: rprintf
    vsnprintf( buffer, sizeof( buffer ), format, args );
    DBG( "libinput " << priority << ": " << chop( buffer, "\n" ) );
};

static const libinput_interface _interface = { _open, _close };

context_t::context_t(
    on_event_t          on_event,
    string_t const &    seat,
    udev_t const &      udev
) :
    OBJECT_T(),
    rep( libinput_udev_create_context( & _interface, this, udev.rep ) ),
    _on_event( on_event ),
    _seat( seat ),
    _thread( * this )
{
    if ( not rep ) {
        ERR( "Failed to initialize libinput context." );
    };
    libinput_log_set_handler( rep, & _log_handler );
    libinput_log_set_priority(
        rep,
        #if defined( DEBUG )
            LIBINPUT_LOG_PRIORITY_DEBUG
        #else
            LIBINPUT_LOG_PRIORITY_INFO
        #endif
    );
};

context_t::~context_t(
) {
    CATCH_ALL(
        disable();
    );
    libinput_unref( rep );
    rep = nullptr;
};

void
context_t::enable(
) {
    int err = libinput_udev_assign_seat( rep, _seat.c_str() );
    if ( err ) {
        ERR( "Failed to assign set to libinput context." );
    };
    _thread.start();
    _state = enabled;
};

void
context_t::disable(
) {
    switch ( _state ) {
        case inited : {
            // Not enabled — nothing to do.
        } break;
        case enabled : {
            /*
                This code relies on the application (since signal disposition is process-wide):
                Application should handle SIGINT signal to let ppoll be interrupted.
            */
            _thread.kill( SIGINT );
            _thread.join();
            _state = disabled;
        } break;
        case disabled : {
            // Already disabled — nothing to do.
        } break;
    };
};

int
context_t::fd(
) {
    return libinput_get_fd( rep );
};

int
context_t::open(
    string_t const &    path,
    int                 flags
) {
    DBG( "Opening file " << q( path ) << "..." );
    int fd = ::open( path.c_str(), flags );
    if ( fd == -1 ) {
        int err = errno;
        ERR( "Failed to open file " << q( path ) << ": " << posix::syserrmsg( err ) << "." );
    };
    _files[ fd ] = path;
    DBG( _files.size() << " files open so far" );
    return fd;
};

void
context_t::close(
    int fd
) {
    DBG( "Closing file " << q( _files[ fd ] ) << "..." );
    int err = ::close( fd );
    if ( err ) {
        err = errno;
        WRN( "Failed to close fd #" << fd << ": " << posix::syserrmsg( err ) << "." );
    };
    _files.erase( fd );
    DBG( _files.size() << " files open so far" );
};

context_t::thread_t::thread_t(
    context_t & context
) :
    _context( context )
{
};

void
context_t::thread_t::body(
) {
    /*
        In the very beginning there are some add_device events in the queue. However, poll does not
        sense them and will wait until the first "real" input event. It is not documented but
        Libinput debug tools (event-debug and list-devices) are implemented in such way.
    */
    struct pollfd pfd {
        .fd      = _context.fd(),
        .events  = POLLIN,
        .revents = 0,
    };
    for ( ; ; ) {
        for ( ; ; ) {
            int err = libinput_dispatch( _context.rep );
            if ( err ) {
                ERR( "Libinput dispatch failed: " << posix::syserrmsg( -err ) << "." );
            };
            event_t event( _context );
            auto type = event.type();
            if ( type == LIBINPUT_EVENT_NONE ) {
                break;
            };
            _context._on_event( event.rep );
        };
        int error = ::poll( & pfd, 1, -1 );
        if ( error < 0 ) {
            error = errno;
            if ( error == EINTR ) { // Interrupted system call.
                break;              // This is not an actual error, just exit the loop.
            };
            ERR( "Failed to poll fd #" << pfd.fd << ": " << posix::syserrmsg( error ) << "." );
        };
    };
};

// -------------------------------------------------------------------------------------------------
// event_t
// -------------------------------------------------------------------------------------------------

context_t::event_t::event_t(
    context_t & context
) :
    OBJECT_T(),
    rep( libinput_get_event( context.rep ) )
{
};

context_t::event_t::~event_t(
) {
    libinput_event_destroy( rep );
};


context_t::event_t::type_t
context_t::event_t::type(
) const {
    if ( rep == nullptr ) {
        return LIBINPUT_EVENT_NONE;
    } else {
        return libinput_event_get_type( rep );
    };
};

std::ostream &
operator <<(
    std::ostream &              stream,
    context_t::event_t::type_t  type
) {
    static std::map< context_t::event_t::type_t, string_t > const map = {
        #define EVENT( event )  { LIBINPUT_EVENT_ ## event, lc( # event ) }
        EVENT( NONE                    ),
        EVENT( DEVICE_ADDED            ),
        EVENT( DEVICE_REMOVED          ),
        EVENT( KEYBOARD_KEY            ),
        EVENT( POINTER_MOTION          ),
        EVENT( POINTER_MOTION_ABSOLUTE ),
        EVENT( POINTER_BUTTON          ),
        EVENT( POINTER_AXIS            ),
        EVENT( TOUCH_DOWN              ),
        EVENT( TOUCH_UP                ),
        EVENT( TOUCH_MOTION            ),
        EVENT( TOUCH_CANCEL            ),
        EVENT( TOUCH_FRAME             ),
        EVENT( TABLET_TOOL_AXIS        ),
        EVENT( TABLET_TOOL_PROXIMITY   ),
        EVENT( TABLET_TOOL_TIP         ),
        EVENT( TABLET_TOOL_BUTTON      ),
        EVENT( TABLET_PAD_BUTTON       ),
        EVENT( TABLET_PAD_RING         ),
        EVENT( TABLET_PAD_STRIP        ),
        EVENT( GESTURE_SWIPE_BEGIN     ),
        EVENT( GESTURE_SWIPE_UPDATE    ),
        EVENT( GESTURE_SWIPE_END       ),
        EVENT( GESTURE_PINCH_BEGIN     ),
        EVENT( GESTURE_PINCH_UPDATE    ),
        EVENT( GESTURE_PINCH_END       ),
        #undef EVENT
    };
    auto it = map.find( type );
    if ( it == map.end() ) {
        stream << "(* unknown event type #" << int( type ) << " *)";
    } else {
        stream << it->second;
    };
    return stream;
};

std::ostream &
operator <<(
    std::ostream &              stream,
    libinput_log_priority       priority
) {
    stream << int( priority );
    return stream;
};

}; // namespace li
}; // namespace tapper

// end of file //
