/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/posix.hpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    POSIX wrappers interface.
**/

#ifndef _POSIX_HPP_
#define _POSIX_HPP_

#include "base.hpp"

#include <initializer_list>
#include <stdexcept>    // std::runtime_error

#include <pthread.h>
#include <signal.h>

namespace tapper {
namespace posix {

    string_t get_env( string_t const & name );
    string_t base_name( string_t const & path );
    string_t syserrmsg( int error );
    void     sleep();

    // ---------------------------------------------------------------------------------------------
    // error_t
    // ---------------------------------------------------------------------------------------------

    class error_t : public std::runtime_error {

        public :

            using std::runtime_error::runtime_error;
            explicit error_t( string_t const & what, int error );

    }; // class error_t

    // ---------------------------------------------------------------------------------------------
    // signal_t
    // ---------------------------------------------------------------------------------------------

    namespace signal {

        /**
            Convenient operations on signal sets.
        **/
        struct set_t {
            explicit set_t();
            explicit set_t( std::initializer_list< int > const & list );
            void  clear();
            set_t & operator +=( int signo );
            sigset_t rep;
        }; // struct set_t

        struct action_t : sigaction {
            typedef void ( * handler_t )( int );
            explicit action_t(
                handler_t     handler = nullptr,
                set_t const & mask    = set_t(),
                int           flags   = 0
            );
        };

        action_t action( int signal, action_t const & action );
        set_t    mask( int how, set_t const & set );
        int      wait( set_t const & set );
        void     kill( pid_t pid, int signal );

    }; // namespace signal

    // ---------------------------------------------------------------------------------------------
    // thread_t
    // ---------------------------------------------------------------------------------------------

    class thread_t : public object_t {

        public :

            explicit thread_t();
            virtual ~thread_t();

            void start();
            void join();
            void kill( int signal );

            static uint_t no();

        protected :

            virtual void body() = 0;

        private :

            static void * _body( void * arg );

            enum state_t {
                inited,
                started,
                joined
            }; // enum state_t

            state_t         _state;
            pthread_t       _handle;

        protected :

            string_t        _what;

    }; // class thread_t

}; // namespace posix
}; // namespace tapper

#endif // _POSIX_HPP_

// end of file //
