/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/settings.cpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
**/

#include "settings.hpp"

#include <iostream>
#include <stdexcept>

#include "reverse.hpp"

#if ENABLE_GNOME
    #include <giomm/settings.h>
    #include "gnome.hpp"
#endif // ENABLE_GNOME

namespace tapper {

/**
    The code below uses macro `ERR`, which throws an exception of type `error_t`.
**/
typedef std::domain_error error_t;

typedef std::map< settings_t::bell_t,        string_t > bell_to_str_t;
typedef std::map< settings_t::interceptor_t, string_t > interceptor_to_str_t;
typedef std::map< settings_t::layouter_t,    string_t > layouter_to_str_t;

typedef reverse_map_t< bell_to_str_t        > str_to_bell_t;
typedef reverse_map_t< interceptor_to_str_t > str_to_interceptor_t;
typedef reverse_map_t< layouter_to_str_t    > str_to_layouter_t;

static bell_to_str_t const bell_to_str {
    //~ { settings_t::bell_t::Default,  "default",  },
    { settings_t::bell_t::disabled, "disabled", },
    { settings_t::bell_t::enabled,  "enabled",  },
};

static interceptor_to_str_t const interceptor_to_str {
    //~ { settings_t::interceptor_t::Default,  "default",  },
    { settings_t::interceptor_t::Auto,     "auto",     },
    { settings_t::interceptor_t::libinput, "libinput", },
    { settings_t::interceptor_t::xrecord,  "xrecord",  },
};

static layouter_to_str_t const layouter_to_str {
    //~ { settings_t::layouter_t::Default,     "default",     },
    { settings_t::layouter_t::Auto,        "auto",        },
    { settings_t::layouter_t::gnome_shell, "gnome-shell", },
    { settings_t::layouter_t::xkb,         "xkb",         },
};

static str_to_bell_t        const str_to_bell        = reverse( bell_to_str        );
static str_to_interceptor_t const str_to_interceptor = reverse( interceptor_to_str );
static str_to_layouter_t    const str_to_layouter    = reverse( layouter_to_str    );

string_t
settings_t::to_string(
    bell_t  bell
) {
    return bell_to_str.at( bell );
};

string_t
settings_t::to_string(
    interceptor_t   interceptor
) {
    return interceptor_to_str.at( interceptor );
};

string_t
settings_t::to_string(
    layouter_t  layouter
) {
    return layouter_to_str.at( layouter );
};

settings_t::interceptor_t
settings_t::to_interceptor(
    string_t const & str
) {
    try {
        return str_to_interceptor.at( str );
    } catch ( std::out_of_range const & ) {
        ERR( "Bad interceptor " << q( str ) << "." );
    };
};

settings_t::layouter_t
settings_t::to_layouter(
    string_t const & str
) {
    try {
        return str_to_layouter.at( str );
    } catch ( std::out_of_range const & ) {
        ERR( "Bad layouter " << q( str ) << "." );
    };
};

/** Returns list of all interceptor names. **/
strings_t
settings_t::interceptors(
) {
    strings_t list;
    for ( auto const & it : interceptor_to_str ) {
        // cppcheck-suppress useStlAlgorithm
        list.push_back( it.second );
    };
    return list;
};

/** Returns list of all layouter names. **/
strings_t
settings_t::layouters(
) {
    strings_t list;
    for ( auto const & it : layouter_to_str ) {
        // cppcheck-suppress useStlAlgorithm
        list.push_back( it.second );
    };
    return list;
};

settings_t::assignments_t
operator +(
    settings_t::assignments_t const & lhs,
    settings_t::assignments_t const & rhs
) {
    auto r = lhs;
    return r += rhs;
};

/** Load settings from GSettings. **/
settings_t
settings_t::load(
) {
    #if ENABLE_GNOME
        settings_t r;
        auto s = Gio::Settings::create( PACKAGE_ID );
        // Options:
        r.interceptor = interceptor_t( s->get_enum( "interceptor" ) );
        r.layouter    = layouter_t(    s->get_enum( "layouter"    ) );
        Glib::Variant< bool > bell;
        auto set = s->get_user_value( "bell", bell );
        if ( set ) {
            r.bell = bell.get() ? bell_t::enabled : bell_t::disabled;
        } else {
            r.bell = bell_t::unset;
        };
        // Mappings:
        Glib::Variant< assignments_t > assignments;
        s->get_value( "assignments", assignments );
        r.assignments = assignments.get();
        return r;
    #else
        ERR( "Program is built without Gnome support." );
    #endif // ENABLE_GNOME
};

/** Save settings to GSettings. **/
void
settings_t::save(
) const {
    #if ENABLE_GNOME
        auto s = Gio::Settings::create( PACKAGE_ID );
        // Options:
        if ( interceptor == interceptor_t::unset ) {
            s->reset( "interceptor" );
        } else {
            static_assert( sizeof( interceptor ) <= sizeof( int ) );
            s->set_enum( "interceptor", int( interceptor ) );
        };
        if ( layouter == layouter_t::unset ) {
            s->reset( "layouter" );
        } else {
            static_assert( sizeof( layouter ) <= sizeof( int ) );
            s->set_enum( "layouter", int( layouter ) );
        };
        if ( bell == bell_t::unset ) {
            s->reset( "bell" );
        } else {
            s->set_boolean( "bell", bell == bell_t::enabled );
        };
        // Assignments:
        s->set_value( "assignments", Glib::Variant< assignments_t >::create( assignments ) );
        // Make sure changes are really applied:
        g_settings_sync(); // glibmm does not have such a function, let's call gio directly.
    #else
        ERR( "Program is built without Gnome support." );
    #endif // ENABLE_GNOME
};

/**
    Merge assignments. Key assignments in the second argument overrides key assignments in the
    first argument. Layout #0 is a special: it means "unassign the key".
**/
settings_t::assignments_t &
operator +=(
    settings_t::assignments_t &       lhs,
    settings_t::assignments_t const & rhs
) {
    for ( auto & assignment : rhs ) {
        if ( assignment.second ) {
            lhs.insert( assignment );
        } else {
            lhs.erase( assignment.first );
        };
    };
    return lhs;
};

/**
    Merge two settings. For example, one settings is loaded from GSettings, another one is built
    from the command line. Non-default options from the second settings overrides corresponding
    options in the first settings. Key assignments are merged too.
**/
settings_t &
operator +=(
    settings_t &        lhs,
    settings_t const &  rhs
) {
    if ( rhs.interceptor != settings_t::interceptor_t::unset ) {
        lhs.interceptor = rhs.interceptor;
    };
    if ( rhs.layouter != settings_t::layouter_t::unset ) {
        lhs.layouter = rhs.layouter;
    };
    if ( rhs.bell != settings_t::bell_t::unset ) {
        lhs.bell = rhs.bell;
    };
    lhs.assignments += rhs.assignments;
    return lhs;
};

/** Merge two settings. **/
settings_t
operator +(
    settings_t const & lhs,
    settings_t const & rhs
) {
    auto r = lhs;
    r += rhs;
    return r;
};

}; // namespace tapper

// end of file //
