/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/strings.cpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    String manipulating utilities implementation.
**/

#include "string.hpp"

#include <algorithm>

namespace tapper {

/**
    Perl-like `split`.
**/
strings_t
split(
    char              chr,  ///< Delimiter.
    string_t const &  str,  ///< String to split.
    size_t            limit ///< Maximal number of strings to return.
) {
    strings_t res;
    size_t size = str.size();
    if ( limit == 0 ) {
        // cppcheck-suppress shadowFunction
        auto q = str.find_last_not_of( chr );
        if ( q == npos ) {
            return res;
        }; // if
        size = q + 1;
    }; // if
    size_t p = 0;
    while ( limit > 1 and p < size ) {
        // cppcheck-suppress shadowFunction
        auto q = str.find( chr, p );
        if ( q == npos ) {
            q = size;
        };
        res.push_back( str.substr( p, q - p ) );
        p = q + 1;
        limit -= 1;
    }; // while
    if ( p <= size ) {
        res.push_back( str.substr( p, size - p ) );
    }; // if
    return res;
}; // split

/**
    Perl-like `join`.
**/
string_t
join(
    string_t const &  joiner,
    strings_t const & strings
) {
    string_t r;
    for ( size_t i = 0, end = strings.size(); i < end; ++ i ) {
        r += ( i == 0 ? "" : joiner );
        r += strings[ i ];
    };
    return r;
};


/**
    Searches a string vector for a string.

    @return Index of the found string, or `npos` if the string is not found.
**/
size_t
find(
    string_t const &  needle,   ///< String to search for.
    strings_t const & haystack  ///< Strings to search in.
) {
    auto it = std::find( haystack.begin(), haystack.end(), needle );
    return it == haystack.end() ? npos : it - haystack.begin();
}; // find

/**
    Extracts specified number of characters from the beginning of a string. If the string is
    shorter than requested number of characters, all characters of the string are returned.

    @return String containing extracted characters.
**/
string_t
left(
    string_t const & str,   ///< String to extract characters from.
    size_t           len    ///< Number of characters to extract.
) {
    return str.substr( 0, len );
}; // left

/**
    Extracts specified number of characters from the end of a string. If the string is shorter than
    requested number of characters, all characters of the string are returned.

    @return String containing extracted characters.
**/
string_t
right(
    string_t const & str,   ///< String to extract characters from.
    size_t           len    ///< Number of characters to extract.
) {
    return str.substr( str.size() >= len ? str.size() - len : 0, len );
}; // right

/**
    Returns `true` if a string looks like an integer number. More precisely, returns `true` if the
    string comprises only digits (no leading or trailing spaces are allowed, no sign is allowed
    too), and `false` otherwise.
**/
bool
like_integer(
    string_t const & str
) {
    if ( str.empty() ) {
        return false;
    };
    size_t pos = ( str[ 0 ] == '+' or str[ 0 ] == '-' ) ? 1 : 0;
    if ( pos >= str.size() ) {
        return false;
    };
    return str.find_first_not_of( "0123456789", pos ) == npos;
}; // like_integer

/**
    Returns an uppercased version of string. If Gnome support is disabled, the function works only
    with ASCII characters.
**/
string_t
uc(
    string_t const & str
) {
    #if ENABLE_GNOME
        return str.uppercase();
    #else
        string_t res;
        res.reserve( str.size() );
        for ( size_t i = 0, end = str.size(); i < end; ++ i ) {
            res += toupper( str[ i ] );
        }; // for
        return res;
    #endif // ENABLE_GNOME
}; // uc

/**
    Returns an uppercased version of string. If Gnome support is disabled, the function works only
    with ASCII characters.
**/
string_t
lc(
    string_t const & str
) {
    #if ENABLE_GNOME
        return str.lowercase();
    #else
        string_t res;
        res.reserve( str.size() );
        for ( size_t i = 0, end = str.size(); i < end; ++ i ) {
            res += tolower( str[ i ] );
        }; // for
        return res;
    #endif // ENABLE_GNOME
}; // lc

string_t
chop(
    string_t const & str,   ///< String to chop suffix from.
    string_t const & sfx    ///< Suffix to chop.
) {
    if ( right( str, sfx.size() ) == sfx ) {
        return left( str, str.size() - sfx.size() );
    } else {
        return str;
    }; // if
}; // chop

string_t
q(
    string_t const & str
) {
    return "`" + str + "'";
};

}; // namespace tapper

// end of file //
