/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/tapper.cpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

#include "tapper.hpp"

namespace tapper {

// -------------------------------------------------------------------------------------------------
// tapper_t
// -------------------------------------------------------------------------------------------------

tapper_t::tapper_t(
    interceptor_t &         interceptor,
    layouter_t &            layouter,
    assignments_t const &   assignments,
    bool                    show_taps
) :
    OBJECT_T(),
    _interceptor( interceptor ),
    _layouter( layouter ),
    _assignments( assignments ),
    _show_taps( show_taps ),
    _autorepeat_timeout( _layouter.get_autorepeat_delay() ),
    _key_range( _interceptor.key_range() ),
    _key_state( uint_t( _key_range.max ) + 1, false )
{
    // TODO: autorepeat_rate
}; // ctor

tapper_t::~tapper_t(
) {
    // TODO: stop interceptor?
}; // dtor

void
tapper_t::start(
) {
    _interceptor.set_on_event( std::bind( & tapper_t::_on_event, this, std::placeholders::_1 ) );
    _interceptor.start();
}; // start

void
tapper_t::stop(
) {
    _interceptor.stop();
}; // stop

void
tapper_t::_on_event(
    event_t const & event
) {
    TRACE();
    assert( event.key );
    if ( _key_range.includes( event.key ) ) {
        if ( event.state == key_state_t::pressed ) {
            if ( _key_state[ event.key ] ) {
                DBG( "k" << event.key << u8"↡ " << _pressed_keys );
                if ( event.key == _last_key ) {
                    // Look like key press is autorepeating.
                    // BTW, we are having autorepeat timeout for free...
                    // TODO: Enable it. Print a log message.
                    //~ master.autorepeat_rate.timeout = time - pressed_at;
                }; // if
                /*
                    No need in updating keyboard state -- we already know the key is pressed.
                    Number of pressed keys is not changed too. But let us reset the last pressed
                    key -- if autorepeating takes place this is not a tap.
                */
                _last_key = 0;
            } else {
                // Update keyboard state:
                DBG( "k" << event.key << u8"↓ " << _pressed_keys );
                _key_state[ event.key ] = true;
                _pressed_keys += 1;
                _last_key      = event.key;
                _pressed_at    = event.time;
            }; // if
        } else {
            DBG( "k" << event.key << u8"↑ " << _pressed_keys );
            /*
                This check is a must because unconditional decreasing `_pressed_keys` may cause
                wrong result. For example, if the program started from the command line, the first
                received event will likely be releasing of Enter key.
            */
            if ( _key_state[ event.key ] ) {
                _key_state[ event.key ] = false;
                assert( _pressed_keys > 0 );
                -- _pressed_keys;
            };
            if (
                event.key == _last_key
                and _pressed_keys == 0
                and event.time - _pressed_at <= _autorepeat_timeout
            ) {
                DBG( "k" << event.key << u8"⇵" );
                _on_tap( event.key );
            }; // if
            _last_key = 0;
        }; // if
    } else {
        WRN(
            "Interceptor reported key " << event.key
                << " out of expected range " << _key_range << "."
        );
    }; // if
};

void
tapper_t::_on_tap(
    key_t  key
) {
    TRACE();
    if ( _show_taps ) {
        OUT( "Key " << _interceptor.key_name( key ) << " tapped." );
    };
    auto it = _assignments.find( key );
    if ( it != _assignments.end() ) {
        DBG( "L" << it->second );
        _layouter.activate( it->second );
    };
};

}; // namespace tapper

// end of file //
