/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/tapper.hpp

    Copyright © 2014, 2016—2017, 2019 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

#ifndef _TAPPER_HPP_
#define _TAPPER_HPP_

#include "base.hpp"
#include "settings.hpp"
#include "interceptor.hpp"
#include "layouter.hpp"

namespace tapper {

// -------------------------------------------------------------------------------------------------
// tapper_t
// -------------------------------------------------------------------------------------------------

/**
    Tapper.

    Tapper listens keyboard events (provided by interceptor) detects taps, and activates keyboard
    layouts (with help from layouter).
**/
class tapper_t : public object_t {

    public :

        typedef settings_t::assignments_t assignments_t;

    public :

        explicit tapper_t(
            interceptor_t &         interceptor,
            layouter_t &            layouter,
            assignments_t const &   assignments,
            bool                    show_taps = false
        );
        ~tapper_t();
        void start();
        void stop();

    private :

        typedef interceptor_t::event_t     event_t;
        typedef interceptor_t::key_state_t key_state_t;

        void _on_event( event_t const & event );
        void _on_tap( key_t key );

    private :

        interceptor_t & _interceptor;
        layouter_t &    _layouter;
        assignments_t   _assignments;
        bool            _show_taps { false };
        time_t          _autorepeat_timeout { 500 };
            // TODO: Update it?

        key_range_t     _key_range;

        /**
            Keyboard state. It maps keys to boolean values. `true` means the corresponding key is
            pressed now, `false` means the key is not pressed.
        **/
        std::vector< bool > _key_state;

        /**
            Number of currently pressed keys. The number of currently pressed keys can be achieved
            from `_key_state`, but scanning the `_key_state` is rather slow. Maintaining number of
            pressed keys is simple and fast.
        **/
        uint_t _pressed_keys { 0 };

        /**
            Key code of the last pressed key or 0. Used to detect taps: if code of released key
            equals to code of the last pressed key, it could be a tap.
        **/
        key_t _last_key { 0 };

        /**
            Time of the last key press event, which code was saved in `last_pressed_key`.
            Meaningful only if `last_pressed_key` is not 0.
        **/
        time_t _pressed_at { 0 };

}; // class tapper_t

}; // namespace tapper

#endif // _TAPPER_HPP_

// end of file //
